# -*- coding: utf-8 -*-

from .exceptions import ClientException, SizeException
from . import debug, cache


class Size:
    """Manage operations related to Droplet size"""

    def __init__(self, render):
        self.render = render

    @cache
    @debug
    def list(self):
        """List Droplet sizes"""
        try:
            return [s for s in self.render('sizes', '/sizes')]
        except ClientException as e:
            raise SizeException(str(e))

    @cache
    @debug
    def id_from_name(self, name):
        """Tranlsate size name into ID"""
        name = name.upper()
        matches = [s.id for s in self.list() if s.name == name]
        if len(matches) == 1:
            return matches[0]
        else:
            raise SizeException('"%s" is not a valid size' % name)

    @cache
    @debug
    def name_from_id(self, id):
        """Translate size ID into human readable value"""
        ret = next((s.name for s in self.list() if s.id == id), None)
        if ret:
            return ret
        raise SizeException("No size found for id %s" % id)



# -*- coding: utf-8 -*-

from .exceptions import ClientException, SnapshotException
from . import debug, cache
from .region import Region


class Snapshot:
    """Manage operations related to Droplet snapshots"""

    def __init__(self, render):
        self.render = render
        self.region = Region(self.render)

    @cache
    @debug
    def list(self):
        """List snapshots"""
        return self.render('images', '/images', params={
            'filter': 'my_images'
        })

    def show(self, name):
        """Show details for a single snapshot"""
        id = self.id_from_name(name)
        return self.render('image', '/images/%s' % id)

    @debug
    def transfer(self, name, region):
        id = self.id_from_name(name)
        region_id = self.region.id_from_name(region)

        try:
            return self.render('event_id', '/images/%s/transfer' % id, params={
                'region_id': region_id
            })
        except ClientException as e:
            raise ImageException(str(e))

    @debug
    def id_from_name(self, name):
        """Translate snapshot name into ID"""
        matches = [s.id for s in self.list() if s.name.lower() == name.lower()]
        if len(matches) == 0:
            raise SnapshotException('No snapshot named "%s" found' % name)
        elif len(matches) == 1:
            return matches[0]
        else:
            raise SnapshotException('More than one match for "%s"' % name)

    @debug
    def destroy(self, name):
        """Destroy snapshot"""
        id = self.id_from_name(name)
        return self.render('event', '/images/%s/destroy' % id)



# -*- coding: utf-8 -*-

from .exceptions import ClientException, DropletException
from . import debug, cache
from .snapshot import Snapshot
from .size import Size
from .region import Region
from .image import Image
from .sshkey import SSHKey


class Droplet:
    """Manage operations related to Droplets."""

    def __init__(self, render):
        self.render = render
        self.snap = Snapshot(self.render)
        self.size = Size(self.render)
        self.region = Region(self.render)
        self.image = Image(self.render)
        self.sshkey = SSHKey(self.render)

    @debug
    def list(self):
        """Return a list of Droplets."""
        try:
            return self.render('droplets', '/droplets')
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def id_from_name(self, name):
        """Tranlate a hostname into its Droplet ID."""
        matches = [d.id for d in self.list() if d.name == name]
        if len(matches) == 0:
            raise DropletException('No Droplet found')
        elif len(matches) == 1:
            return matches[0]
        else:
            raise DropletException('More than one Droplet matches "%s"' % (
                                   name))

    @debug
    def name_from_id(self, id):
        """Translate a Droplet ID into its hostname."""
        try:
            return self.render('droplet', '/droplets/%s' % id).name
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def show(self, name):
        """Retrieve information about a single Droplet."""
        id = next((d.id for d in self.list() if d.name == name), None)

        try:
            return self.render('droplet', '/droplets/%s' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def create(self, name=None, size=None, image=None,
               region=None, keys=None, disable_virtio=False,
               private_networking=False):
        """
        Create a Droplet. Returns a :class:`Struct <Struct>` object.

        :param name: Hostname for Droplet.
        :param size: Size of Droplet (512MB, 1GB...)
        :param image: Image to use for Droplet.
        :param region: Region to boot Droplet in.
        :param keys: List of registered SSH keys to associate with Droplet.
        :param disable_virtio: Disable VirtIO for Droplet (not recommended).
        :param private_networking: Add a private IP (if available).
        """

        if (name and size and image and region) is None:
            raise DropletException("Name, size, image and region "
                                   "are all required.")

        virtio = False if disable_virtio else True

        if name in [d.name for d in self.list()]:
            raise DropletException(
                'This would create two droplets with the same hostname.')

        size_id = self.size.id_from_name(size)
        image_id = self.image.id_from_name(image)
        region_id = self.region.id_from_name(region)
        ssh_key_ids = [
            k.id for k in self.sshkey.list() if k.name.lower() in [
                n for n in keys]]

        params = {
            'name': name,
            'size_id': size_id,
            'image_id': image_id,
            'region_id': region_id,
        }

        if ssh_key_ids:
            params['ssh_key_ids'] = ','.join([str(k) for k in ssh_key_ids])
        if virtio:
            params['virtio'] = 1
        if private_networking:
            params['private_networking'] = 1

        try:
            return self.render('droplet', '/droplets/new', params=params)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def start(self, name):
        """Boot Droplet"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/power_on' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def shutdown(self, name):
        """Send ACPI shutdown signal to Droplet"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/shutdown' % id,
                               method='POST')
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def snapshot(self, name, snapshot_name):
        """Create a snapshot of the Droplet (must be shut down first)"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/snapshot' % id,
                               params={'name': snapshot_name})
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def restore(self, droplet_name, snapshot_name):
        """Restore Droplet from a snapshot (must be shut down first)"""

        droplet_id = self.id_from_name(droplet_name)
        snapshot_id = self.snap.id_from_name(snapshot_name)

        try:
            return self.render('event_id', '/droplets/%s/restore' % (
                droplet_id), params={'image_id': snapshot_id})
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def rebuild(self, droplet_name, image_name):
        """Rebuild Droplet from a stock image (must be shut down first)"""
        droplet_id = self.id_from_name(droplet_name)
        image_id = self.image.id_from_name(image_name)

        return self.render('event_id', '/droplets/%s/rebuild' % (droplet_id),
                           params={'image_id': image_id})

    @debug
    def rename(self, frm, to):
        """Rename Droplet (must be shut down first)"""
        from_id = self.id_from_name(frm)
        if to not in [d.name for d in self.list()]:
            try:
                return self.render('event_id', '/droplets/%s/rename' % from_id,
                                   params={'name': to})
            except ClientException as e:
                raise DropletException(str(e))
        raise DropletException("There is already a Droplet named '%s'" % to)

    @debug
    def resize(self, name, size):
        """Resize Droplet (must be shut down first)"""
        id = self.id_from_name(name)
        size_id = self.size.id_from_name(size)
        try:
            return self.render('event_id', '/droplets/%s/resize' % id,
                               params={'size_id': size_id})
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def destroy(self, name, no_scrub=False):
        """Destroy Droplet"""
        id = self.id_from_name(name)
        scrub_data = False if no_scrub else True
        try:
            return self.render('event_id', '/droplets/%s/destroy' % id,
                               params={'scrub_data': 1 if scrub_data else 0})
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def status(self, name):
        """Retrive Droplet status"""
        id = self.id_from_name(name)
        try:
            return self.render('droplet', '/droplets/%s' % id).status
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def reboot(self, name):
        """Reboot Droplet"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/reboot' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def powercycle(self, name):
        """Power cycle Droplet"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/power_cycle' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def poweroff(self, name):
        """Power down Droplet"""
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/power_off' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def backups(self, action, name):
        """
        Enable or disable backups for Droplet

        :param action: either 'enable' or 'disable'
        :param name: Name of Droplet
        """
        id = self.id_from_name(name)
        try:
            if action == 'enable':
                return self.render('event_id',
                                   '/droplets/%s/enable_backups' % id)
            elif action == 'disable':
                return self.render('event_id',
                                   '/droplets/%s/disable_backups' % id)
        except ClientException as e:
            raise DropletException(str(e))

    @debug
    def passwordreset(self, name):
        """
        Reset root password for Droplet
        (results in email to registered Digital Ocean account)
        """
        id = self.id_from_name(name)
        try:
            return self.render('event_id', '/droplets/%s/password_reset' % id)
        except ClientException as e:
            raise DropletException(str(e))



# -*- coding: utf-8 -*-

__name__ = 'pontoon'
__description__ = 'A Python CLI for Digital Ocean'
__version__ = '0.1.2'
__author__ = 'Ross Duggan'
__author_email__ = 'ross.duggan@acm.org'
__url__ = 'https://github.com/duggan/pontoon'
__copyright__ = 'Copyright Ross Duggan 2013'

from .cache import cache
from .log import debug

from .pontoon import Pontoon
from .command import Command
from .pontoon import ClientException
from .exceptions import *



# -*- coding: utf-8 -*-

from .exceptions import ClientException, SSHKeyException
from . import debug, cache


class SSHKey:
    """Manage operations related to SSH keys"""

    def __init__(self, render):
        self.render = render

    @cache
    def list(self):
        """List registered SSH keys"""
        try:
            return [s for s in self.render('ssh_keys', '/ssh_keys')]
        except ClientException as e:
            raise SSHKeyException(str(e))

    @debug
    def show(self, name):
        """Retrieve SSH key details (including public portion of key)"""
        id = self.id_from_name(name)
        try:
            return self.render('ssh_key', '/ssh_keys/%s' % id)
        except ClientException as e:
            raise SSHKeyException(str(e))

    @debug
    def add(self, name, public_key):
        """Register a public key with your Digital Ocean account"""
        if name not in [s.name for s in self.list()]:
            try:
                return self.render('ssh_key', '/ssh_keys/new', params={
                    'name': name,
                    'ssh_pub_key': public_key
                })
            except ClientException as e:
                raise SSHKeyException(str(e))
        raise SSHKeyException("Aborted: this would create two "
                              "keys with the same name.")

    @debug
    def replace(self, name, public_key):
        """Replace a registered SSH key"""
        id = self.id_from_name(name)
        params = {
            'ssh_pub_key': public_key
        }
        try:
            return self.render('ssh_key', '/ssh_keys/%s/edit' % id,
                               params=params)
        except ClientException as e:
            raise SSHKeyException(str(e))

    @debug
    def destroy(self, name):
        """Deresgister an SSH key"""
        id = self.id_from_name(name)
        try:
            return self.render('status', '/ssh_keys/%s/destroy' % id)
        except ClientException as e:
            raise SSHKeyException(str(e))

    @cache
    @debug
    def id_from_name(self, name):
        """Tranlsate an SSH key name into its ID"""
        ret = next((r.id for r in self.list() if r.name == name), None)
        if ret:
            return ret
        raise SSHKeyException("No key found called %s" % name)

    @cache
    @debug
    def name_from_id(self, id):
        """Translate an SSH key id into its name"""
        ret = next((r.name for r in self.list() if r.id == id), None)
        if ret:
            return ret
        raise SSHKeyException("No key found for id %s" % id)



# -*- coding: utf-8 -*-

import re
import sys
import contextlib
from random import randrange
from datetime import datetime, timedelta

from .exceptions import ClientException

# Python 2/3 compatibility for capture_stdout
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class Data(object):
    pass


def timestamp(hours=0):
    """Mocked Digital Ocean timestamp"""
    return (datetime.utcnow() + timedelta(
        hours=hours)).strftime("%Y-%m-%dT%H:%M:%SZ")


def get_builtins():
    """Python 2.x and 3.x have different names for accessing builtins"""
    try:
        __import__('builtins')
        return 'builtins'
    except ImportError:
        return '__builtin__'


def _raise(ex=None):
    """Wrapper for exceptions so they can be thrown from inside lambdas"""
    if ex:
        exception.append(ex)
    if len(exception):
        raise exception.pop()


def event_response():
    return {
        'event_id': randrange(9999),
        'status': 'OK',
    }


@contextlib.contextmanager
def capture_stdout():
    """Captures STDOUT and turns it into an object"""
    old = sys.stdout
    capturer = StringIO()
    sys.stdout = capturer
    data = Data()
    yield data
    sys.stdout = old
    data.result = capturer.getvalue()


exception = []

mocked = {
    'droplets': [
        {
            'id': 1, 'name': 'foo', 'image_id': 1, 'region_id': 2,
            'backups_active': False, 'status': 'active',
            'ip_address': '192.0.2.1', 'size_id': 3, 'locked': False,
            'created_at': timestamp(-200), 'private_ip_address': None,
        },
        {
            'id': 2, 'name': 'bar', 'image_id': 1, 'region_id': 2,
            'backups_active': False, 'status': 'active',
            'ip_address': '192.0.2.2', 'size_id': 2, 'locked': False,
            'created_at': timestamp(-5), 'private_ip_address': None,
        },
        {
            'id': 3, 'name': 'baz', 'image_id': 1, 'region_id': 1,
            'backups_active': False, 'status': 'active',
            'ip_address': '192.0.2.3', 'size_id': 1, 'locked': False,
            'created_at': timestamp(), 'private_ip_address': None,
        },
    ],
    'regions': [
        {
            'id': 1, 'name': 'Foo York 1',
        },
        {
            'id': 2, 'name': 'Bardam 1',
        },
        {
            'id': 3, 'name': 'Foo Barbaz 1',
        },
        {
            'id': 4, 'name': 'Foo York 2',
        },
    ],
    'images': [
        {
            'id': 1, 'name': 'Foobuntu 12.04 x64', 'distribution': 'Foobuntu',
        },
        {
            'id': 2, 'name': 'Foobuntu 12.04 x32', 'distribution': 'Foobuntu',
        },
        {
            'id': 3, 'name': 'Bar 6.0 x64', 'distribution': 'Bar',
        },
        {
            'id': 4, 'name': 'Bar 6.0 x32', 'distribution': 'Bar',
        },
    ],
    'snapshots': [
        {
            'id': 1024, 'name': 'snapshot-foo',
            'distribution': 'Foobuntu',
        },
        {
            'id': 2048, 'name': 'snapshot-bar-2013-10-10',
            'distribution': 'Foobuntu',
        },
        {
            'id': 4096, 'name': 'snapshot-baz-pre-install',
            'distribution': 'Foobuntu',
        },
    ],
    'sizes': [
        {
            'id': 1, 'name': '512MB',
        },
        {
            'id': 2, 'name': '1GB',
        },
        {
            'id': 3, 'name': '2GB',
        },
    ],
    'ssh_keys': [
        {
            'id': 1, 'name': 'foobarbaz',
            'ssh_pub_key': 'ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAIEA1on8gxCGJJWSRT'
                           '4uOrR13mUaUk0hRf4RzxSZ1zRbYYFw8pfGesIFoEuVth4HKyF8'
                           'k1y4mRUnYHP1XNMNMJl1JcEArC2asV8sHf6zSPVffozZ5TT4Sf'
                           'sUu/iKy9lUcCfXzwre4WWZSXXcPff+EHtWshahu3WzBdnGxm5X'
                           'oi89zcE= test@example.com'
        },
    ],
    'events': [
        {
            'id': 999, 'action_status': 'done', 'droplet_id': 1,
            'event_type_id': 1, 'percentage': 100,
        },
        {
            'id': 888, 'action_status': 'done', 'droplet_id': 2,
            'event_type_id': 2, 'percentage': 100,
        },
        {
            'id': 777, 'action_status': 'done', 'droplet_id': 3,
            'event_type_id': 3, 'percentage': 100,
        },
    ],
}


def _respond(target, method='GET', params={}):
    """Fake out responses from Digital Ocean."""

    if len(params) > 0:
        if re.match("^/images$", target):
            if 'filter' in params:
                return {'images': (mocked.get('snapshots') +
                                   mocked.get('images'))}

        if re.match("^/images/\d+/transfer$", target):
            if 'region_id' in params:
                return event_response()

        elif re.match("^/droplets/\d+/resize$", target):
            if 'size_id' in params:
                return event_response()

        elif re.match("^/droplets/\d+/snapshot$", target):
            if 'name' in params:
                return event_response()

        elif re.match("^/droplets/\d+/rebuild$", target):
            if 'image_id' in params:
                return event_response()

        elif re.match("^/droplets/\d+/rename$", target):
            if 'name' in params:
                return event_response()

        elif re.match("^/droplets/\d+/restore$", target):
            if 'image_id' in params:
                return event_response()

        elif re.match("^/droplets/new$", target):
            new = {
                'id': randrange(100, 999),
                'name': params['name'],
                'image_id': params['image_id'],
                'region_id': params['region_id'],
                'backups_active': False,
                'status': 'new',
                'ip_address': '192.0.2.%s' % randrange(10, 255),
                'size_id': params['size_id'],
                'locked': False,
                'created_at': timestamp(),
                'private_ip_address': None
                }
            mocked['droplets'].append(new)

            return {'droplet': new}

        elif re.match("^/ssh_keys/new$", target):
            new = {
                'id': randrange(100, 999),
                'name': params['name'],
                'ssh_pub_key': params['ssh_pub_key']
                }
            mocked['ssh_keys'].append(new)

            return {'ssh_key': new}

        elif re.match("^/ssh_keys/\d+/edit$", target):
            r = re.match("^/ssh_keys/(\d+)/edit$", target)
            id = int(r.group(1))
            replaced = []
            new = {
                'id': id,
                'ssh_pub_key': params['ssh_pub_key']
            }
            for key in mocked.get('ssh_keys'):
                if key['id'] == id:
                    new['name'] = key['name']
                    replaced.append(new)
                else:
                    replaced.append(key)
            mocked['ssh_keys'] = replaced

            return {'ssh_key': next((
                k for k in mocked.get(
                    'ssh_keys') if k['id'] == id), None)}

        elif re.match("^/droplets/\d+/destroy$", target):
            if 'scrub_data' in params:
                return event_response()
    else:
        options = {
            '/droplets': {'droplets': mocked.get('droplets')},
            '/regions': {'regions': mocked.get('regions')},
            '/images': {'images': (mocked.get('images') +
                                   mocked.get('snapshots'))},
            '/sizes': {'sizes': mocked.get('sizes')},
            '/ssh_keys': {'ssh_keys': mocked.get('ssh_keys')},
            '/droplets/\d+/reboot': event_response(),
            '/droplets/(\d+)': lambda x: {'droplet': next(
                (d for d in mocked.get('droplets') if d['id'] == x), None)},
            '/droplets/\d+/power_cycle': event_response(),
            '/droplets/\d+/power_off': event_response(),
            '/droplets/\d+/power_on': event_response(),
            '/droplets/\d+/enable_backups': event_response(),
            '/droplets/\d+/disable_backups': event_response(),
            '/droplets/\d+/shutdown': event_response(),
            '/droplets/\d+/password_reset': event_response(),
            '/images/(\d+)': lambda x: {'image': next(
                (d for d in (mocked.get('images') +
                             mocked.get('snapshots')) if d['id'] == x), None)},
            '/images/\d+/destroy': {'event': event_response()},
            '/events/(\d+)': lambda x: {'event': next(
                (d for d in mocked.get('events') if d['id'] == x), None)},
            '/ssh_keys/(\d+)': lambda x: {'ssh_key': next(
                (d for d in mocked.get('ssh_keys') if d['id'] == x), None)},
            '/ssh_keys/\d+/destroy': event_response(),

        }
        for pattern, func in options.items():
            r = re.match("^%s$" % pattern, target)
            if r:
                try:
                    arg = r.group(1)
                    res = func(int(arg))
                    if not next(iter(res.values())):
                        raise ClientException("A Mock Error")
                    return res
                except IndexError:
                    pass
                return func



# -*- coding: utf-8 -*-

from . import configure, ui
from . import Pontoon


class Command:
    """
    Base class for CLI commands.

    Parses docstrings and initalizes client.
    """

    def __init__(self, doc, argv=None):
        from docopt import docopt

        configure.logger()
        conf = configure.combined()

        pontoon = Pontoon(conf['client_id'], conf['api_key'],
                          mock=configure.mock_mode)
        args = docopt(doc, argv=argv)

        self.pontoon = pontoon
        self.args = args

    def run(self, override=None):
        """Run the command specified by the docstring.

        :param override: Run a command other than the one specified.
        """

        if override is None:
            command = next(
                (k for k, v in self.args.items()
                    if k in dir(self) and v), None)
        else:
            command = override

        try:
            return getattr(self, command)()
        except (KeyboardInterrupt, EOFError):
            return 0
        except KeyError:
            ui.message("No command '%s'" % command)
        return 1



# -*- coding: utf-8 -*-

from .exceptions import ClientException, EventException
from . import debug, cache


class Event:
    """Manage operations related to events"""

    def __init__(self, render):
        self.render = render

    @debug
    def show(self, id):
        """Show details for a single event"""
        try:
            return self.render('event', '/events/%s' % id)
        except ClientException as e:
            raise EventException(str(e))

    @debug
    def type_from_id(self, id):
        """Tranlate event type IDs into something human readable."""
        # There should be an API endpoint for this
        # but let's pretend there is until there is.
        types = {
            8: 'snapshot',
        }
        if id in types:
            return types[id]
        return "unknown (%s)" % id



# -*- coding: utf-8 -*-

from .exceptions import ClientException, ImageException
from . import debug, cache


class Image:
    """Manage operations related to images."""

    def __init__(self, render):
        self.render = render

    @cache
    @debug
    def list(self):
        """List available images."""
        try:
            return [s for s in self.render('images', '/images')]
        except ClientException as e:
            raise ImageException(str(e))

    def show(self, name):
        """Show details for a single image."""
        id = self.id_from_name(name)
        try:
            return self.render('image', '/images/%s' % id)
        except ClientException as e:
            raise ImageException(str(e))

    @debug
    def oses(self):
        """Return a list of the Operating System flavours made availble
        through the various images."""
        oses = []
        for image in self.list():
            oses.append(image.distribution)
        return [i for i in set(oses)]

    @cache
    @debug
    def id_from_name(self, name):
        """Translate image name into an ID"""
        res = next((
            r.id for r in self.list() if r.name.lower() == name.lower()),
            None)
        if res:
            return res
        raise ImageException("No image called %s" % name)

    @cache
    @debug
    def name_from_id(self, id):
        """Translate image ID into a name"""
        try:
            res = self.render('image', '/images/%s' % id).name
            if res:
                return res
        except ClientException:
            pass
        raise ImageException("No image found for id %s" % id)



# -*- coding: utf-8 -*-

from . import debug
from .exceptions import ClientException
from .droplet import Droplet
from .event import Event
from .image import Image
from .region import Region
from .size import Size
from .snapshot import Snapshot
from .sshkey import SSHKey
from .mocking import _respond


API_HOST = 'api.digitalocean.com'
API_PORT = 80


class Struct:
    """
    A generic object for encapsulating response data.

    Accepts keyword arguments, translates them into attributes on the object.
    """
    def __init__(self, **entries):
        self.__dict__.update(entries)


class Pontoon:
    """A user-created :class:`Pontoon <Pontoon>` object.

    Central interface to various pontoon objects.

    :param client_id: Digital Ocean API Client ID.
    :param api_key: Digital Ocean API key.
    :param host: HTTP URL for Digital Ocean API.
    :param port: Port to access host on.
    :param secure: Access API over HTTPS.
    :param mock: Use mocked request interface.

    Usage::

        >>> from pontoon import Pontoon
        >>> p = Pontoon('my_client_id', 'my_api_key')
        >>> p.droplet.list()
        [<pontoon.pontoon.Struct instance at 0x106ecf950>]
    """
    def __init__(self, client_id, api_key,
                 host=API_HOST, port=API_PORT, secure=True, mock=False):

        self.__mock = mock
        self.__client_id = client_id
        self.__api_key = api_key
        self.__port = port
        self.__host = host
        self.__secure = secure

        self.droplet = Droplet(self.render)
        self.event = Event(self.render)
        self.image = Image(self.render)
        self.region = Region(self.render)
        self.size = Size(self.render)
        self.snapshot = Snapshot(self.render)
        self.sshkey = SSHKey(self.render)

    @debug
    def render(self, key, path, method='GET', params={}):
        """
        Translates dictionary responses into
        :class:`Struct <Struct>` objects.

        :param key: top level key expected in a valid response.
        :param path: path fragment for API resource.
        :param method: HTTP verb for request.
        :param params: Parameters for request.
        """
        if self.__mock:
            self.request = _respond

        resp = self.request(path, method=method, params=params)

        if not resp:
            raise ClientException("Problem with request: "
                                  "%s, %s" % (path, params))
        content = resp.get(key, None)
        if type(content).__name__ == 'list':
            return [Struct(**r) for r in content]
        elif type(content).__name__ == 'dict':
            return Struct(**content)
        elif type(content).__name__ in ['str', 'int']:
            return content
        raise ClientException("Malformed response: %s" % content)

    @debug
    def request(self, path, method='GET', params={}):
        """Prepare and proxy to requests module."""
        import requests

        if method not in ['GET', 'POST']:
            raise ClientException("%s method not supported" % method)
        headers = {
            'User-Agent': 'pontoon/client'
        }

        params['client_id'] = self.__client_id
        params['api_key'] = self.__api_key

        url = self.get_url(path)

        if method == 'POST':
            headers['Content-Type'] = "application/json"
            response = requests.post(url, headers=headers, params=params)
        else:
            response = requests.get(url, headers=headers, params=params)

        if response.status_code == 200:
            if response.json():
                json = response.json()
                error_msg = json.get('error_message', None)
                if error_msg:
                    raise ClientException(error_msg)
                return json
            raise ClientException('Empty json!')
        elif response.status_code == 401:
            raise ClientException('Access Denied')
        elif response.status_Code == 404:
            raise ClientException('Not Found')

        raise ClientException(('Status code: %d, full response: %s' %
                              (response.status_code, response.json())))

    def get_url(self, path):
        """Tranlsate path into a fully qualified URL."""
        port = "" if self.__port == 80 else ":%d" % self.__port
        protocol = "https://" if self.__secure else "http://"
        return "%s%s%s%s" % (protocol, self.__host, port, path)



# -*- coding: utf-8 -*-

"""
Taken from https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
"""

import logging
import functools

logger = logging.getLogger(__name__)


def cache(obj):
    """Cache the given object hashed on arguments"""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        logline = (obj.__module__, obj.__name__, args)
        if key not in cache:
            logger.debug("caching %s.%s %s" % logline)
            cache[key] = obj(*args, **kwargs)
        logger.debug("returning cached %s.%s %s" % logline)
        return cache[key]
    return memoizer



# -*- coding: utf-8 -*-

from __future__ import print_function

import readline
import textwrap
from os.path import (isfile, expanduser,
                     basename, splitext, join)
from time import sleep
from sys import stdout

# Python 2.6 compatibility
try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict

# Python 2/3 compatibility
try:
    user_input = raw_input
except:
    user_input = input


def ticker():
    """A loading/waiting indicator.

    Sends a '.' to the screen, resets, and sleeps.
    """
    stdout.write('.')
    stdout.flush()
    sleep(1)


def ask_yesno(question):
    """Present a string as a yes/no question on an interactive prompt"""
    question = "%s (y/n)" % question
    if user_input("%-15s: " % question).strip().lower() == 'y':
        return True
    return False


def ask(question):
    """Present a question with freeform input on an interactive prompt"""
    question += ':'
    response = user_input("%-15s " % question)
    return response


def valid_path(path):
    """Check whether a given string resolves to a path on the filesystem"""
    return isfile(expanduser(path))


def full_path(path):
    """Expand a ~/ prefixed path to a full path"""
    return join(expanduser(path))


def filename_from_path(path):
    """Get a filename from a given path"""
    return splitext(basename(expanduser(path)))[0]


def mask(text, masker='*'):
    """Hide part of a string,

    If the input is small, hide the entire string.
    Otherwise, show only the last few characters.
    """
    length = len(text)
    coverage = length / 100.0 * 20
    if (coverage < 0.9):
        result = (masker * length)
        return result
    masked_length = int(length - coverage)
    result = "%s%s" % ((masker * masked_length), text[masked_length:])
    return result


def format_droplet_info(machine, size="", region="", image=""):
    """Present Droplet information in a more human parseable format"""
    d = OrderedDict()
    d['id'] = machine.id
    d['name'] = machine.name
    d['size'] = size
    d['image'] = image
    d['region'] = region

    del machine.region_id
    del machine.image_id
    del machine.size_id

    for k, v in machine.__dict__.items():
        d[k] = v

    return d


def format_event(event, type="", droplet=""):
    """Present event information in a more human parseable format"""
    e = OrderedDict()
    e['id'] = event.id
    e['type'] = type
    e['droplet'] = droplet

    del event.event_type_id
    del event.droplet_id

    for k, v in event.__dict__.items():
        e[k] = v

    return e


def message(text):
    """Wrapper for the `print` function"""
    print(text)
    return text


def heading(text, boxwidth=60):
    """Create a 'heading' styled textbox"""
    return box(text, decor="-", decor_y="|", boxwidth=boxwidth)


def notify(text, boxwidth=60):
    """Create a 'notification' styled textbox"""
    return box(text, decor="*", boxwidth=boxwidth)


def warning(text, boxwidth=60):
    """Create a 'warning' styled textbox"""
    return box(text, decor="!", boxwidth=boxwidth)


def box(text, decor='*', decor_x=None, decor_y=None,
        boxwidth=60, borderwidth=2):
    """Create a formatted textbox for highlighting important information"""
    decor_x = decor_x if decor_x else decor
    decor_y = decor_y if decor_y else decor
    decor_x_width = len(decor_x) if len(decor_x) > 0 else 1
    decor_y_width = len(decor_y) if len(decor_y) > 0 else 1
    decor_x_multiplier = int(boxwidth / decor_x_width)
    decor_y_multiplier = decor_y_width * 2

    textwidth = boxwidth - decor_y_multiplier - (borderwidth * 2)
    text = textwrap.wrap(text, textwidth)
    text = "\n".join([
        "{decor_y}{text}{decor_y}".format(
            decor_y=decor_y, text=m.center(
                boxwidth-decor_y_multiplier)) for m in text])
    border_x = (decor_x * decor_x_multiplier)
    border_space = "{decor_y}{spacer}{decor_y}".format(
        decor_y=decor_y, spacer=(" " * (boxwidth-decor_y_multiplier)))
    spacing = (' ' * boxwidth)
    text = "\n".join([
        spacing, border_x, border_space, text,
        border_space, border_x, spacing])
    print(text)
    return text



# -*- coding: utf-8 -*-

from .exceptions import ClientException, RegionException
from . import debug, cache


class Region:
    """Manage operations related to regions."""

    def __init__(self, render):
        self.render = render

    @cache
    @debug
    def list(self):
        """List regions"""
        try:
            return [s for s in self.render('regions', '/regions')]
        except ClientException as e:
            raise RegionException(str(e))

    @cache
    @debug
    def id_from_name(self, name):
        """Translate region name into ID"""
        ret = next((
            r.id for r in self.list() if r.name.lower() == name.lower()),
            None)
        if ret:
            return ret
        raise RegionException("No region called %s" % name)

    @cache
    @debug
    def name_from_id(self, id):
        """Translate region ID into name"""
        ret = next((r.name for r in self.list() if r.id == id), None)
        if ret:
            return ret
        raise RegionException("No region with id %s" % id)



# -*- coding: utf-8 -*-

from __future__ import print_function

import sys
import os
import io
from subprocess import CalledProcessError
from pytest import raises
import pytest
from mock import MagicMock, patch


# Put lib dir into path so can be tested without installing
lib_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
test_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, lib_dir)

from docopt import DocoptExit
from pontoon import (PontoonException, DropletException,
                     SnapshotException, ConfigureException,
                     SSHKeyException, ImageException,
                     RegionException, SizeException,
                     EventException)
from pontoon import Pontoon
from pontoon import Command
from pontoon import ClientException
from pontoon.pontoon import Struct
from pontoon import ui
from pontoon import configure
from pontoon.mocking import (_raise, capture_stdout, event_response,
                             get_builtins, _respond, mocked)


_request = MagicMock(name='request')
_request.side_effect = _respond


class Data(object):
    pass


_input = MagicMock(name='_input')


_expanduser = MagicMock(name='expanduser')


_sleep = MagicMock(name='sleep')


_call = MagicMock(name='call')


_Popen = MagicMock(name='Popen')


_open = MagicMock(name='open')


_getuid = MagicMock(name='getuid')
_getuid.return_value = 1


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestRender:

    client = Pontoon('foo', 'bar')

    def test_render(self):
        with raises(ClientException):
            self.client.render('foo', "/foo/%s/bar" % Struct)

        with raises(ClientException):
            self.client.render('foo', '/foo/100/bar')


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestDroplet(object):

    droplet = Pontoon('foo', 'bar').droplet

    def test_list(self):
        _request.side_effect = _respond
        for server in self.droplet.list():
            assert isinstance(server, Struct)

    def test_id_from_name(self):
        _request.side_effect = _respond

        assert self.droplet.id_from_name('foo') == 1

        with raises(DropletException):
            self.droplet.id_from_name('not-a-droplet')

        mocked['droplets'].append({
            'id': 9,
            'name': 'baz',
        })
        with raises(DropletException):
            self.droplet.id_from_name('baz')

    def test_name_from_id(self):
        _request.side_effect = _respond
        assert self.droplet.name_from_id(2) == 'bar'

        with raises(DropletException):
            self.droplet.name_from_id(999)

    def test_show(self):
        _request.side_effect = _respond

        server = self.droplet.show("foo")
        assert server.id == 1
        server = self.droplet.show("baz")
        assert server.id == 3
        server = self.droplet.show("bar")
        assert server.id == 2

        with raises(DropletException):
            self.droplet.show(200)

    def test_create(self):
        _request.side_effect = _respond
        result = self.droplet.create(name="newfoo", size="512MB",
                                     image="Bar 6.0 x64", region="Bardam 1",
                                     keys=["foobarbaz"])
        assert result.size_id == 1
        assert result.image_id == 3
        assert result.region_id == 2

        with raises(DropletException):
            result = self.droplet.create(name="newfoo", size="512MB",
                                         image="Bar 6.0 x64",
                                         region="Bardam 1",
                                         keys=["foobarbaz"])

        with raises(DropletException):
            self.droplet.create()

    def test_start(self):
        _request.side_effect = _respond
        e = self.droplet.start('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.start('non-existant')

    def test_shutdown(self):
        _request.side_effect = _respond
        e = self.droplet.shutdown('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.shutdown('non-existant')

    def test_snapshot(self):
        _request.side_effect = _respond
        e = self.droplet.snapshot('foo', 'snapshot-foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.snapshot('non-existant', 'snapshot-foo')

    def test_restore(self):
        _request.side_effect = _respond
        e = self.droplet.restore('foo', 'snapshot-foo')
        assert isinstance(e, int)

        with raises(SnapshotException):
            self.droplet.restore('foo', 'not-snapshot')

    def test_rebuild(self):
        _request.side_effect = _respond
        e = self.droplet.rebuild('foo', 'Foobuntu 12.04 x64')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.rebuild('non-existant', 'not-snapshot')

        with raises(ImageException):
            self.droplet.rebuild('foo', 'not-snapshot')

    def test_rename(self):
        _request.side_effect = _respond
        e = self.droplet.rename('foo', 'foofoo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.rename('foo', 'bar')

    def test_resize(self):
        _request.side_effect = _respond
        e = self.droplet.resize('foo', '1GB')
        assert isinstance(e, int)

        with raises(SizeException):
            self.droplet.resize('foo', '64MB')

    def test_destroy(self):
        _request.side_effect = _respond
        e = self.droplet.destroy('foo')
        assert isinstance(e, int)
        e = self.droplet.destroy('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.destroy('non-existant')

    def test_status(self):
        _request.side_effect = _respond
        status = self.droplet.status('foo')
        assert status == 'active'

        with raises(DropletException):
            self.droplet.status('non-existant')

    def test_reboot(self):
        _request.side_effect = _respond
        e = self.droplet.reboot('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.reboot('non-existant')

    def test_powercycle(self):
        _request.side_effect = _respond
        e = self.droplet.powercycle('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.powercycle('non-existant')

    def test_poweroff(self):
        _request.side_effect = _respond
        e = self.droplet.poweroff('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.poweroff('non-existant')

    def test_backups(self):
        _request.side_effect = _respond
        e = self.droplet.backups('enable', 'foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.backups('enable', 'non-existant')

        e = self.droplet.backups('disable', 'foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.backups('disable', 'non-existant')

    def test_passwordreset(self):
        _request.side_effect = _respond
        e = self.droplet.passwordreset('foo')
        assert isinstance(e, int)

        with raises(DropletException):
            self.droplet.passwordreset('non-existant')


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestImage(object):

    image = Pontoon('foo', 'bar').image

    def test_list(self):
        for img in self.image.list():
            assert isinstance(img, Struct)

    def test_show(self):
        img = self.image.show("Foobuntu 12.04 x64")
        assert img.id == 1

    def test_oses(self):
        assert sorted(['Bar', 'Foobuntu']) == sorted(self.image.oses())

    def test_id_from_name(self):
        assert 1 == self.image.id_from_name('Foobuntu 12.04 x64')
        with raises(ImageException):
            foo = self.image.id_from_name('Foo')

    def test_name_from_id(self):
        assert 'Foobuntu 12.04 x32' == self.image.name_from_id(2)
        with raises(ImageException):
            foo = self.image.name_from_id(10)


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestSnapshot(object):

    snapshot = Pontoon('foo', 'bar').snapshot
    snapshot_list = MagicMock('snapshot_list')
    some_snapshot = Data()
    some_snapshot.id = 2
    some_snapshot.name = 'snapshot-bar'
    snapshot_list.return_value = [some_snapshot, some_snapshot]

    def test_list(self):
        for snap in self.snapshot.list():
            assert isinstance(snap, Struct)

    def test_show(self):
        img = self.snapshot.show("snapshot-foo")
        assert img.id == 1024

    def test_id_from_name(self):
        assert 1024 == self.snapshot.id_from_name('snapshot-foo')
        with raises(SnapshotException):
            foo = self.snapshot.id_from_name('not-a-snapshot')

    @patch('pontoon.snapshot.Snapshot.list', snapshot_list)
    def test_id_from_name_two(self):
        with raises(SnapshotException):
            foo = self.snapshot.id_from_name('snapshot-bar')

    def test_destroy(self):
        e = self.snapshot.destroy('snapshot-foo')
        assert e.status == 'OK'
        with raises(SnapshotException):
            foo = self.snapshot.destroy('not-a-snapshot')

    def test_transfer(self):
        e = self.snapshot.transfer('snapshot-foo', 'Bardam 1')
        assert isinstance(e, int)


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestEvent(object):

    event = Pontoon('foo', 'bar').event

    def test_show(self):
        e = self.event.show(999)
        assert e.action_status == 'done'

        with raises(EventException):
            e = self.event.show(444)

    def test_type_from_id(self):
        t = self.event.type_from_id(8)
        assert t == 'snapshot'

        t = self.event.type_from_id(10)
        assert t == 'unknown (10)'


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestSize(object):

    size = Pontoon('foo', 'bar').size

    def test_list(self):
        for s in self.size.list():
            assert isinstance(s, Struct)

    def test_id_from_name(self):
        assert 1 == self.size.id_from_name('512MB')
        with raises(SizeException):
            self.size.id_from_name('Nonexistant')

    def test_name_from_id(self):
        assert '2GB' == self.size.name_from_id(3)
        with raises(SizeException):
            self.size.name_from_id(99)


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestRegion(object):

    region = Pontoon('foo', 'bar').region

    def test_list(self):
        for r in self.region.list():
            assert isinstance(r, Struct)

    def test_id_from_name(self):
        assert 1 == self.region.id_from_name('Foo York 1')
        with raises(RegionException):
            self.region.id_from_name('Nonexistant')

    def test_name_from_id(self):
        assert 'Bardam 1' == self.region.name_from_id(2)
        with raises(RegionException):
            self.region.name_from_id(99)


@patch('pontoon.pontoon.Pontoon.request', _request)
class TestSSHKey(object):

    sshkey = Pontoon('foo', 'bar').sshkey

    def test_list(self):
        for s in self.sshkey.list():
            assert isinstance(s, Struct)

    def test_show(self):
        key = self.sshkey.show("foobarbaz")
        assert key.name == 'foobarbaz'

    def test_add(self):
        key = self.sshkey.add("barbaz",
                              "this_is_a_poor_attempt_at_ssh_public_key")
        assert key.name == 'barbaz'

        with raises(SSHKeyException):
            key = self.sshkey.add("foobarbaz",
                                  "this_is_a_poor_attempt_at_ssh_public_key")

    def test_replace(self):
        key = self.sshkey.replace("foobarbaz",
                                  "this_is_a_poor_attempt_at_ssh_public_key")
        assert key.name == 'foobarbaz'

    def test_destroy(self):
        e = self.sshkey.destroy("foobarbaz")
        assert e == 'OK'

    def test_id_from_name(self):
        assert 1 == self.sshkey.id_from_name('foobarbaz')
        with raises(SSHKeyException):
            self.sshkey.id_from_name('Nonexistant')

    def test_name_from_id(self):
        assert 'foobarbaz' == self.sshkey.name_from_id(1)
        with raises(SSHKeyException):
            self.sshkey.name_from_id(99)


@patch('pontoon.pontoon.Pontoon.request', _request)
@patch('pontoon.ui.sleep', _sleep)
@patch('pontoon.ui.user_input', _input)
class TestUI:

    thisfile = os.path.realpath(__file__)

    def test_ticker(self):
        with capture_stdout() as capture:
            ui.ticker()
        assert capture.result == ""

    def test_ask_yesno(self):
        _input.return_value = "y"
        assert ui.ask_yesno("Foo?") is True
        _input.return_value = "n"
        assert ui.ask_yesno("Bar?") is False

    def test_ask(self):
        _input.return_value = "bar"
        assert ui.ask("Foo") == 'bar'

    def test_valid_path(self):
        assert ui.valid_path(self.thisfile)

    def test_full_path(self):
        assert ui.full_path(self.thisfile) == self.thisfile

    def test_filename_from_path(self):
        assert ui.filename_from_path(self.thisfile) == 'test_models'

    def test_mask(self):
        assert ui.mask('tiny') == '****'
        assert ui.mask('somewhatlonger') == '***********ger'
        assert (ui.mask('reallylongperhapsabighashorsomethinglikethat') ==
               ('***********************************glikethat'))

    def test_format_droplet_info(self):
        pontoon = Pontoon('foo', 'bar')
        resp = ui.format_droplet_info(pontoon.droplet.show('foo'),
                                      size="512MB", region="Bardam 1",
                                      image="Foobuntu 12.04 x64")
        assert resp['id'] == 1
        assert resp['name'] == 'foo'
        assert resp['region'] == 'Bardam 1'

    def test_format_event(self):
        pontoon = Pontoon('foo', 'bar')
        resp = ui.format_event(pontoon.event.show(999),
                               type=8,
                               droplet='foo')
        assert resp['id'] == 999
        assert resp['type'] == 8
        assert resp['droplet'] == 'foo'

    def test_message(self):
        with capture_stdout() as capture:
            ui.message("foo bar baz")
        assert capture.result == "foo bar baz\n"

    def test_heading(self):
        heading = (
            "------------------------------------------------------------\n"
            "|                                                          |\n"
            "|                           foo                            |\n"
            "|                                                          |\n"
            "------------------------------------------------------------\n"
        )
        with capture_stdout() as capture:
            ui.heading("foo")
        assert heading.strip() == capture.result.strip()

    def test_notify(self):
        notify = (
            "************************************************************\n"
            "*                                                          *\n"
            "*                           foo                            *\n"
            "*                                                          *\n"
            "************************************************************"
        )
        with capture_stdout() as capture:
            ui.notify("foo")
        assert notify.strip() == capture.result.strip()

    def test_warning(self):
        warning = (
            "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
            "!                                                          !\n"
            "!                           foo                            !\n"
            "!                                                          !\n"
            "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
        )
        with capture_stdout() as capture:
            ui.warning("foo")
        assert warning.strip() == capture.result.strip()


@patch('pontoon.configure.getuid', _getuid)
@patch('pontoon.configure.expanduser', _expanduser)
@patch('pontoon.configure.call', _call)
@patch('pontoon.configure.Popen', _Popen)
@patch('%s.open' % get_builtins(), _open)
@patch('pontoon.pontoon.Pontoon.request', _request)
class TestConfigure:

    def test_ssh_tools(self):
        _Popen.side_effect = ["/usr/bin/ssh-keygen", "/usr/bin/openssl"]
        assert configure.ssh_tools() is True

        _Popen.side_effect = lambda *args, **kwargs: (
            _raise(CalledProcessError("foo", "bar")))
        assert configure.ssh_tools() is False

    def test_regenerate_public_key(self):
        _call.return_value = None
        assert configure.regenerate_public_key("foo", "bar") is None

    def test_rsa_keygen(self):
        _call.side_effect = lambda x: print("generating a public key" % x)
        with capture_stdout() as capture:
            configure.rsa_keygen("foo")
        assert capture.result == "generating a public key\n"

    def test_combined(self):
        fake_config = (
            "client_id: foo\n"
            "api_key: bar\n"
            "auth_key: ~/.ssh/foo\n"
            "auth_key_name: foo"
        ).encode('UTF-8')

        _open.side_effect = lambda *args, **kwargs: io.BytesIO(fake_config)
        assert configure.combined() == {
            'client_id': 'foo',
            'api_key': 'bar',
            'auth_key': '~/.ssh/foo',
            'auth_key_name': 'foo',
            'username': 'root',
            'scrub_data': True,

        }

    def test_list_keys(self):
        r = configure.list_keys({'client_id': 'foo', 'api_key': 'bar'})
        assert 'foobarbaz' in [k.name for k in r]

    def test_register_key(self):
        r = configure.register_key({'client_id': 'foo', 'api_key': 'bar'},
                                   'bar',
                                   'bazbazbazbazbazbazbaz')
        assert r is None

        with raises(ConfigureException):
            configure.register_key({'client_id': 'foo', 'api_key': 'bar'},
                                   'foobarbaz',
                                   'bazbazbazbazbazbazbaz')

    def test_read_key(self):
        key_response = 'bazbazbazbazbazbazbaz'.encode('UTF-8')
        _open.side_effect = lambda *args, **kwargs: io.BytesIO(key_response)
        key = configure.read_key('/foo/bar/baz')
        assert key == 'bazbazbazbazbazbazbaz'.encode('UTF-8')

    def test_create_config(self):
        _open.side_effect = lambda *args, **kwargs: io.BytesIO()
        assert configure.create_config({'foo': 'bar', 'baz': 'boo'}) is None
        _open.side_effect = lambda *args, **kwargs: io.BytesIO()
        _getuid.return_value = 0
        assert configure.create_config({'foo': 'bar', 'baz': 'boo'}) is None

    def test_read_config(self):
        fake_config = (
            "client_id: foo\n"
            "api_key: bar\n"
            "auth_key: ~/.ssh/foo\n"
            "auth_key_name: foo"
        ).encode('UTF-8')

        _open.side_effect = lambda *args, **kwargs: io.BytesIO(fake_config)
        assert configure.read_config() == {
            'client_id': 'foo',
            'api_key': 'bar',
            'auth_key': '~/.ssh/foo',
            'auth_key_name': 'foo',
        }
        fake_config = (
            "client_id: foo\n"
            "api_key: bar\n"
            "auth_key: ~/.ssh/foo\n"
        ).encode('UTF-8')

        _open.side_effect = lambda *args, **kwargs: io.BytesIO(fake_config)
        assert configure.read_config() == {
            'client_id': 'foo',
            'api_key': 'bar',
            'auth_key': '~/.ssh/foo',
            'auth_key_name': None,
        }
        _open.side_effect = lambda x: (_raise(IOError))
        configure.read_config() == {
            'client_id': 'foo',
            'api_key': 'bar',
            'auth_key': '~/.ssh/foo',
            'auth_key_name': None,
        }


class TestCommand:

    def test_command(self):
        doc = """Usage: command foo"""
        argv = ['foo']

        class SmallCommand(Command):

            def foo(self):
                print("foo-answer")

            def bar(self):
                print("bar-answer")

        command = SmallCommand(doc, argv=argv)

        with capture_stdout() as capture:
            output = command.run()
        assert capture.result == "foo-answer\n"

        with capture_stdout() as capture:
            output = command.run("bar")
        assert capture.result == "bar-answer\n"

        with raises(DocoptExit):
            command = SmallCommand(doc, argv=['baz'])
